# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Module defining the class for a composited sequence of coordinate transformations.

This module contains the class :class:`~astropy.coordinates.CompositeTransform` that
combines a sequence of transformations into a single transformation. The class has the
same API as a single-step transformation, so it can be used interchangeably with a
single-step transformation.
"""

from astropy.coordinates.transformations.affine import (
    AffineTransform,
    BaseAffineTransform,
    DynamicMatrixTransform,
    StaticMatrixTransform,
)
from astropy.coordinates.transformations.base import CoordinateTransform
from astropy.coordinates.transformations.function import (
    FunctionTransformWithFiniteDifference,
)

__all__ = ["CompositeTransform"]


class CompositeTransform(CoordinateTransform):
    """
    A transformation constructed by combining together a series of single-step
    transformations.

    Note that the intermediate frame objects are constructed using any frame
    attributes in ``toframe`` or ``fromframe`` that overlap with the intermediate
    frame (``toframe`` favored over ``fromframe`` if there's a conflict).  Any frame
    attributes that are not present use the defaults.

    Parameters
    ----------
    transforms : sequence of `~astropy.coordinates.CoordinateTransform` object
        The sequence of transformations to apply.
    fromsys : class
        The coordinate frame class to start from.
    tosys : class
        The coordinate frame class to transform into.
    priority : float or int
        The priority if this transform when finding the shortest
        coordinate transform path - large numbers are lower priorities.
    register_graph : `~astropy.coordinates.TransformGraph` or None
        A graph to register this transformation with on creation, or
        `None` to leave it unregistered.
    collapse_static_mats : bool
        If `True`, consecutive `~astropy.coordinates.StaticMatrixTransform`
        will be collapsed into a single transformation to speed up the
        calculation.

    """

    def __init__(
        self,
        transforms,
        fromsys,
        tosys,
        priority=1,
        register_graph=None,
        collapse_static_mats=True,
    ):
        super().__init__(
            fromsys, tosys, priority=priority, register_graph=register_graph
        )

        if collapse_static_mats:
            transforms = self._combine_statics(transforms)

        self.transforms = tuple(transforms)

    def _combine_statics(self, transforms):
        """
        Combines together sequences of StaticMatrixTransform's into a single
        transform and returns it.
        """
        newtrans = []
        for currtrans in transforms:
            lasttrans = newtrans[-1] if len(newtrans) > 0 else None

            if isinstance(lasttrans, StaticMatrixTransform) and isinstance(
                currtrans, StaticMatrixTransform
            ):
                newtrans[-1] = StaticMatrixTransform(
                    currtrans.matrix @ lasttrans.matrix,
                    lasttrans.fromsys,
                    currtrans.tosys,
                )
            else:
                newtrans.append(currtrans)
        return newtrans

    def __call__(self, fromcoord, toframe):
        curr_coord = fromcoord
        for t in self.transforms:
            # build an intermediate frame with attributes taken from either
            # `toframe`, or if not there, `fromcoord`, or if not there, use
            # the defaults
            # TODO: caching this information when creating the transform may
            # speed things up a lot
            frattrs = {}
            for inter_frame_attr_nm in t.tosys.frame_attributes:
                if hasattr(toframe, inter_frame_attr_nm):
                    attr = getattr(toframe, inter_frame_attr_nm)
                    frattrs[inter_frame_attr_nm] = attr
                elif hasattr(fromcoord, inter_frame_attr_nm):
                    attr = getattr(fromcoord, inter_frame_attr_nm)
                    frattrs[inter_frame_attr_nm] = attr

            curr_toframe = t.tosys(**frattrs)
            curr_coord = t(curr_coord, curr_toframe)

        # this is safe even in the case where self.transforms is empty, because
        # coordinate objects are immutable, so copying is not needed
        return curr_coord

    def _as_single_transform(self):
        """
        Return an encapsulated version of the composite transform so that it appears to
        be a single transform.

        The returned transform internally calls the constituent transforms.  If all of
        the transforms are affine, the merged transform is
        `~astropy.coordinates.DynamicMatrixTransform` (if there are no
        origin shifts) or `~astropy.coordinates.AffineTransform`
        (otherwise).  If at least one of the transforms is not affine, the merged
        transform is
        `~astropy.coordinates.FunctionTransformWithFiniteDifference`.
        """
        # Create a list of the transforms including flattening any constituent CompositeTransform
        transforms = [
            t if not isinstance(t, CompositeTransform) else t._as_single_transform()
            for t in self.transforms
        ]

        if all(isinstance(t, BaseAffineTransform) for t in transforms):
            # Check if there may be an origin shift
            fixed_origin = all(
                isinstance(t, (StaticMatrixTransform, DynamicMatrixTransform))
                for t in transforms
            )

            # Dynamically define the transformation function
            def single_transform(from_coo, to_frame):
                if from_coo.is_equivalent_frame(to_frame):  # loopback to the same frame
                    return None if fixed_origin else (None, None)

                # Create a merged attribute dictionary for any intermediate frames
                # For any attributes shared by the "from"/"to" frames, the "to" frame takes
                #   precedence because this is the same choice implemented in __call__()
                merged_attr = {
                    name: getattr(from_coo, name) for name in from_coo.frame_attributes
                }
                merged_attr.update(
                    {
                        name: getattr(to_frame, name)
                        for name in to_frame.frame_attributes
                    }
                )

                affine_params = (None, None)
                # Step through each transform step (frame A -> frame B)
                for i, t in enumerate(transforms):
                    # Extract the relevant attributes for frame A
                    if i == 0:
                        # If frame A is actually the initial frame, preserve its attributes
                        a_attr = {
                            name: getattr(from_coo, name)
                            for name in from_coo.frame_attributes
                        }
                    else:
                        a_attr = {
                            k: v
                            for k, v in merged_attr.items()
                            if k in t.fromsys.frame_attributes
                        }

                    # Extract the relevant attributes for frame B
                    b_attr = {
                        k: v
                        for k, v in merged_attr.items()
                        if k in t.tosys.frame_attributes
                    }

                    # Obtain the affine parameters for the transform
                    # Note that we insert some dummy data into frame A because the transformation
                    #   machinery requires there to be data present.  Removing that limitation
                    #   is a possible TODO, but some care would need to be taken because some affine
                    #   transforms have branching code depending on the presence of differentials.
                    next_affine_params = t._affine_params(
                        t.fromsys(from_coo.data, **a_attr), t.tosys(**b_attr)
                    )

                    # Combine the affine parameters with the running set
                    affine_params = _combine_affine_params(
                        affine_params, next_affine_params
                    )

                # If there is no origin shift, return only the matrix
                return affine_params[0] if fixed_origin else affine_params

            # The return type depends on whether there is any origin shift
            transform_type = DynamicMatrixTransform if fixed_origin else AffineTransform
        else:
            # Dynamically define the transformation function
            def single_transform(from_coo, to_frame):
                if from_coo.is_equivalent_frame(to_frame):  # loopback to the same frame
                    return to_frame.realize_frame(from_coo.data)
                return self(from_coo, to_frame)

            transform_type = FunctionTransformWithFiniteDifference

        return transform_type(
            single_transform, self.fromsys, self.tosys, priority=self.priority
        )


def _combine_affine_params(params, next_params):
    """
    Combine two sets of affine parameters.

    The parameters for an affine transformation are a 3 x 3 Cartesian
    transformation matrix and a displacement vector, which can include an
    attached velocity.  Either type of parameter can be ``None``.
    """
    M, vec = params
    next_M, next_vec = next_params

    # Multiply the transformation matrices if they both exist
    if M is not None and next_M is not None:
        new_M = next_M @ M
    else:
        new_M = M if M is not None else next_M

    if vec is not None:
        # Transform the first displacement vector by the second transformation matrix
        if next_M is not None:
            vec = vec.transform(next_M)

        # Calculate the new displacement vector
        if next_vec is not None:
            if "s" in vec.differentials and "s" in next_vec.differentials:
                # Adding vectors with velocities takes more steps
                # TODO: Add support in representation.py
                new_vec_velocity = vec.differentials["s"] + next_vec.differentials["s"]
                new_vec = vec.without_differentials() + next_vec.without_differentials()
                new_vec = new_vec.with_differentials({"s": new_vec_velocity})
            else:
                new_vec = vec + next_vec
        else:
            new_vec = vec
    else:
        new_vec = next_vec

    return new_M, new_vec



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Backend implementations for `~astropy.table.Table` interface
with Jupyter notebooks.

"""

__all__ = ["classic", "ipydatagrid"]


# NOTE: The actual deprecation warning is emitted in show_in_notebook
# method in table.py module.
def classic(
    table,
    tableid=None,
    css=None,
    display_length=50,
    table_class="astropy-default",
    show_row_index="idx",
):
    """Render the table in HTML and show it in the Jupyter notebook.

    .. deprecated:: 6.1
       Use :func:`ipydatagrid` instead.

    Parameters
    ----------
    table : `~astropy.table.Table`
        Table to render.
    tableid : str or None
        An html ID tag for the table.  Default is ``table{id}-XXX``, where
        id is the unique integer id of the table object, id(table), and XXX
        is a random number to avoid conflicts when printing the same table
        multiple times.
    table_class : str or None
        A string with a list of HTML classes used to style the table.
        The special default string ('astropy-default') means that the string
        will be retrieved from the configuration item
        ``astropy.table.default_notebook_table_class``. Note that these
        table classes may make use of bootstrap, as this is loaded with the
        notebook.  See `this page <https://getbootstrap.com/css/#tables>`_
        for the list of classes.
    css : str
        A valid CSS string declaring the formatting for the table. Defaults
        to ``astropy.table.jsviewer.DEFAULT_CSS_NB``.
    display_length : int, optional
        Number or rows to show. Defaults to 50.
    show_row_index : str or False
        If this does not evaluate to False, a column with the given name
        will be added to the version of the table that gets displayed.
        This new column shows the index of the row in the table itself,
        even when the displayed table is re-sorted by another column. Note
        that if a column with this name already exists, this option will be
        ignored. Defaults to "idx".

    Returns
    -------
    html : object
        An ``IPython.display.HTML`` instance representing the given table.

    Notes
    -----
    Currently, unlike :meth:`~astropy.table.Table.show_in_browser`
    (with ``jsviewer=True``), this
    method needs to access online Javascript code repositories.  This is due
    to modern browsers' limitations on accessing local files.  Hence, if you
    call this method while offline (and don't have a cached version of
    jquery and jquery.dataTables), you will not get the jsviewer features.

    """
    import numpy as np
    from IPython.display import HTML

    from . import conf
    from .jsviewer import JSViewer

    if tableid is None:
        tableid = f"table{id(table)}-{np.random.randint(1, 1e6)}"

    jsv = JSViewer(display_length=display_length)
    if show_row_index:
        display_table = table._make_index_row_display_table(show_row_index)
    else:
        display_table = table
    if table_class == "astropy-default":
        table_class = conf.default_notebook_table_class
    html = display_table._base_repr_(
        html=True,
        max_width=-1,
        tableid=tableid,
        max_lines=-1,
        show_dtype=False,
        tableclass=table_class,
    )

    columns = display_table.columns.values()
    sortable_columns = [
        i for i, col in enumerate(columns) if col.info.dtype.kind in "iufc"
    ]
    html += jsv.ipynb(tableid, css=css, sort_columns=sortable_columns)
    return HTML(html)


def ipydatagrid(table, **kwargs):
    """Render the table in HTML with ``ipydatagrid`` and show it in
    the Jupyter notebook.

    This function creates an ``ipydatagrid.DataGrid`` object by converting the input
    ``table`` to a ``pandas.DataFrame`` and passing ``**kwargs`` to the constructor.
    The available ``DataGrid`` options can be seen in a Jupyter notebook with
    ``help(ipydatagrid.DataGrid)``.

    .. note::
        This function requires optional dependencies ``pandas`` and ``ipydatagrid``.

    Parameters
    ----------
    table : `~astropy.table.Table`
        Table to render.

    **kwargs : dict, optional
        Keyword arguments accepted by ``ipydatagrid.DataGrid``.

    Returns
    -------
    dg : object
        An ``ipydatagrid.DataGrid`` instance representing the given table.

    """
    from ipydatagrid import DataGrid

    return DataGrid(table.to_pandas(), **kwargs)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest

# test helper.run_tests function
from astropy import test as run_tests


# run_tests should raise ValueError when asked to run on a module it can't find
def test_module_not_found():
    with pytest.raises(ValueError), pytest.deprecated_call(match="The test runner"):
        run_tests(package="fake.module")


# run_tests should raise ValueError when passed an invalid pastebin= option
def test_pastebin_keyword():
    with pytest.raises(ValueError), pytest.deprecated_call(match="The test runner"):
        run_tests(pastebin="not_an_option")
